unit frmmain;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, LResources, Forms, Controls, Graphics, Dialogs,
  StdCtrls, Menus, ActnList, fpcddb, ExtCtrls, ComCtrls;

type

  { TMainForm }

  TMainForm = class(TForm)
    AConfig: TAction;
    AQuit: TAction;
    AQueryFile: TAction;
    AQueryCD: TAction;
    ALMain: TActionList;
    EYear: TEdit;
    EPerformer: TEdit;
    ETitle: TEdit;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    LETitle: TLabel;
    LVTracks: TListView;
    MComments: TMemo;
    MenuItem1: TMenuItem;
    MIQuit: TMenuItem;
    MIConfigure: TMenuItem;
    MIReadCD: TMenuItem;
    MIReadFile: TMenuItem;
    MFile: TMenuItem;
    MMain: TMainMenu;
    ODFile: TOpenDialog;
    PDisk: TPanel;
    SBMain: TStatusBar;
    SPTop: TSplitter;
    procedure AConfigExecute(Sender: TObject);
    procedure AQueryCDExecute(Sender: TObject);
    procedure AQueryFileExecute(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    { private declarations }
    FCDDevice : String;
    FAppName : String;
    FUser : String;
    FHostUrl : String;
    FUseCache : Boolean;
    FCacheDir : String;
    FCDDB : TCDDBParser;
    function DoCDDBCmd(CMD: String; Response: TStream): Boolean;
    function GetCategory(AQuery: String): String;
    function GetDiskContent(const ADiscID, ACategory: String; Content: TStream ): Boolean;
    procedure LoadConfig;
    function LoadFromCache(ADiscID: Integer): Boolean;
    procedure QueryCD(const ADevice: String);
    Function QueryFile(const AFileName: String) : Integer;
    procedure SaveConfig;
    procedure SaveToCache(ADiscID: String; M: TStream);
    function SelectMatch(AMatches : TCDDBQueryMatches) : Integer;
    Function ShowCDDB(AStream: TStream; WithHeader: Boolean = True) : Integer;
    procedure ShowConfig;
    procedure ShowDisk(ADisk: TCDDisk);
    procedure ShowTrack(LI: TListItem; ATrack: TCDTrack);
    function TryCreateCacheDir: Boolean;
  public
    { public declarations }
  end; 

var
  MainForm: TMainForm;

implementation

uses cdrom,discid,inifiles,httpsend,frmselectmatch,frmconfig;

Const
  SConfig     = 'Config';
  KeyDevice   = 'Device';
  KeyUser     = 'User';
  KeyHost     = 'Host';
  KeyCacheDir = 'CacheDir';
  KeyUseCache = 'UseCache';

  Version = '1.0';

  DefaultHost = 'http://freedb.org/~cddb/cddb.cgi';
  SHello    = 'hello=%s+localhost+%s+%s';
  SCmdQuery = 'cmd=cddb+query+%s';
  SCmdRead  = 'cmd=cddb+read+%s+%s';
  SProtocol = 'proto=1';

Resourcestring
  SConfigCDROm       = 'Please configure the CD-ROM device first';
  SNoConfigDir       = 'Could not create the configuration directory.';
  SErrFailedToReadCD = 'Failed to read CD-ROM device %s';
  SErrQueryFailed    = 'Failed to query the CDDB server';
  SErrCDDB           = 'Error in CDDB query: %s';
  SErrCreateCacheFailed = 'Failed to create the cache directory: %s';

{ TMainForm }

procedure TMainForm.AQueryFileExecute(Sender: TObject);
begin
  If ODFile.Execute Then
    QueryFile(ODFile.FileName)
end;

procedure TMainForm.AQueryCDExecute(Sender: TObject);
begin
  If (FCDDevice='') then
    begin
    ShowMessage(SConfigCDROM);
    ShowConfig;
    end;
  If (FCDDevice<>'') then
    QueryCD(FCDDevice);
end;

procedure TMainForm.AConfigExecute(Sender: TObject);
begin
  ShowConfig;
end;

procedure TMainForm.FormCreate(Sender: TObject);

begin
  FCDDB:=TCDDBParser.Create(Self);
  LoadConfig;
  FAppName:=ExtractFIleName(Paramstr(0));
  If (FCDDevice='') then
    begin
    FCDDevice:='/dev/sr0';
    FUser:='anonymous';
    end;
end;

procedure TMainForm.LoadConfig;

Var
  D,FN : String;
  Ini : TIniFile;

begin
  D:=IncludeTrailingPathDelimiter(GetAppConfigDir(False)+'cddb');
  FN:=GetAppConfigFile(False);
  If not ForceDirectories(ExtractFilePath(FN)) then
    begin
    ShowMessage(SNoConfigDir);
    Exit;
    end;
  Ini:=TIniFile.Create(FN);
  try
    FCDDevice:=Ini.ReadString(SConfig,KeyDevice,'');
    FUser:=Ini.ReadString(SConfig,KeyUser,'');
    FHostURL:=Ini.ReadString(SConfig,KeyUser,DefaultHost);
    FuseCache:=Ini.ReadBool(SConfig,KeyUseCache,False);
    FCacheDir:=Ini.ReadString(SConfig,KeyCacheDir,D);
  finally
    Ini.Free;
  end;
end;

procedure TMainForm.QueryCD(Const ADevice : String);

Var
  TheDiscID,Tracks, i : Integer;
  Entries   : Array[1..100] of TTocEntry;
  Q,URL,Category     : String;
  M : TMemoryStream;

begin
  Tracks := ReadCDTOC(ADevice,Entries);
  If (Tracks<=0) then
    begin
    ShowMessage(Format(SErrFailedToReadCD,[ADevice]));
    Exit;
    end;
  TheDiscID:=CDDBDiscID(Entries,Tracks);
  If FUseCache then
    If LoadFromCache(TheDiscID) then
      Exit;
  Q:=GetCDDBQueryString(Entries,Tracks);
  Category:=GetCategory(Q);
  If (Category='') then
    exit;
  M:=TMemoryStream.Create;
  try
    If GetDiskContent(DiscIDToStr(TheDiscID),Category,M) then
      begin
      ShowCDDB(M);
      If FUseCache then
        SaveToCache(DiscIDToStr(TheDiscID),M);
      end;
  finally
    M.Free;
  end;
end;

Function TMainForm.GetDiskContent(Const ADiscID,ACategory : String; Content : TStream) : Boolean;

begin
  Result:=DoCDDBCmd(Format(SCmdRead,[ACategory,ADiscID]),Content);
end;

Function TMainForm.DoCDDBCmd(CMD : String; Response : TStream) : Boolean;

Var
  Http : THTTPSend;
  U,URL : String;

begin
  Result:=False;
  U:=Fuser;
  If U='' then
    U:='Anonymous';
  HTTP:=THTTPSend.Create;
  Try
    Url:=FHostURL+'?';
    URL:=URL+StringReplace(Cmd,' ','+',[rfReplaceAll]);
    URL:=URL+'&'+Format(SHello,[U,FAppName,Version]);
    URL:=URL+'&proto=1';
    Result:=HTTP.HTTPMethod('GET',URL);
    If Not Result then
      ShowMessage(SErrQueryFailed)
    else
      begin
      Response.CopyFrom(HTTP.Document,0);
      Response.Position:=0;
      end;
  Finally
    HTTP.Free;
  end;
end;

Function TMainForm.GetCategory(AQuery : String) : String;

Var
  S : TMemoryStream;
  M : TCDDBQueryMatches;
  I : integer;

begin
  Result:='';
  M:=TCDDBQueryMatches.Create(TCDDBQueryMatch);
  try
    S:=TMemoryStream.Create;
    try
      If not DoCDDBCmd(Format(SCMdQuery,[AQuery]),S) then
        Exit;
      I:=FCDDB.ParseCDDBQueryResponse(S,M,True);
    finally
      S.Free;
    end;
    I:=SelectMatch(M);
    If I<>-1 then
      Result:=M[i].Category;
  finally
    M.Free;
  end;
end;

Function TMainForm.SelectMatch(AMatches : TCDDBQueryMatches) :Integer;


begin
  If AMatches.Count=0 then
    Result:=-1
  else If AMatches.Count=1 then
    Result:=0
  else
    With TSelectMatchForm.Create(Self) do
      try
        Matches:=AMatches;
        If ShowModal=mrOK then
          Result:=Selected
        else
          Result:=-1;
      finally
        Free;
      end;
end;

Function TMainForm.QueryFile(Const AFileName : String) : Integer;

Var
  F : TFileStream;

begin
  F:=TFileStream.Create(AFileName,fmOpenRead);
  try
    Result:=ShowCDDB(F,False);
  finally
    F.Free;
  end;
end;

Function TMainForm.ShowCDDB(AStream : TStream; WithHeader: Boolean = True): Integer;


begin
  Result:=FCDDB.ParseCDDBReadResponse(AStream,WithHeader);
  If (Result>=1) then
    ShowDisk(FCDDB.Disks[0]);
end;

procedure TMainForm.ShowDisk(ADisk : TCDDisk);

Var
  I  : Integer;
  LI : TListItem;
  T  : TCDTrack;

begin
  ETitle.Text:=ADisk.Title;
  EPerformer.Text:=ADisk.Performer;
  MComments.Lines.Text:=ADisk.Extra;
  If (ADisk.Year<>0) then
    EYear.Text:=IntToStr(ADisk.Year)
  else
    EYear.Clear;
  With LVTracks do
    begin
    BeginUpdate;
    try
      Items.Clear;
      For I:=0 to ADisk.Tracks.Count-1 do
        begin
        Li:=Items.Add;
        T:=ADisk.Tracks[i];
        Li.Data:=T;
        ShowTrack(LI,T);
        end;
    finally
      EndUpdate;
    end;
    end;
end;

procedure TMainForm.ShowTrack(LI : TListItem; ATrack : TCDTrack);

begin
  LVTracks.BeginUpdate;
  try
    LI.Caption:=IntToStr(ATrack.Index+1);
    LI.SubItems.Clear;
    LI.SubItems.Add(ATrack.Title);
    If (ATrack.Duration<>0) then
      LI.SubItems.Add(FormatDateTime('nn:ss',ATrack.Duration))
    else
      LI.SubItems.Add('');
    LI.SubItems.Add(ATrack.Performer);
    LI.SubItems.Add(ATrack.Extra);
  finally
    LVTracks.EndUpdate;
  end;
end;

procedure TMainForm.ShowConfig;

begin
  With TConfigurationForm.Create(Self) do
    try
      CDDevice:=FCDDevice;
      URL:=FHostURL;
      UserName:=FUser;
      UseCache:=FUseCache;
      CacheDir:=FCacheDir;
      If (ShowModal=mrOK) then
        begin
        FCDDevice := CDDevice;
        FHostURL  := URL;
        FUser     := UserName;
        FUseCache := UseCache;
        FCacheDir := CacheDir;
        SaveConfig;
        end;
    finally
      Free;
    end;
end;

procedure TMainForm.SaveConfig;

Var
  FN : String;
  Ini : TIniFile;

begin
  FN:=GetAppConfigFile(False);
  If not ForceDirectories(ExtractFilePath(FN)) then
    begin
    ShowMessage(SNoConfigDir);
    Exit;
    end;
  Ini:=TIniFile.Create(FN);
  With Ini Do
  try
    WriteString(SConfig,KeyDevice,FCDDevice);
    WriteString(SConfig,KeyUser,FUser);
    WriteString(SConfig,KeyUser,FHostURL);
    WriteBool(SConfig,KeyUseCache,FuseCache);
    WriteString(SConfig,KeyCacheDir,FCacheDir);
  finally
    Ini.Free;
  end;
  If FUseCache then
    TryCreateCacheDir;
end;

Function TMainForm.TryCreateCacheDir : Boolean;

begin
  Result:=ForceDirectories(FCacheDir);
  If Not Result then
    ShowMessage(Format(SErrCreateCacheFailed,[FCacheDir]));
end;

Procedure TMainForm.SaveToCache(ADiscID : String; M : TStream);

Var
  L : TStringList;
  FN : String;

begin
  If not TryCreateCacheDir then
    exit;
  L:=TStringList.Create;
  try
    M.Position:=0;
    L.LoadFromStream(M);
    // Delete command result;
    L.Delete(0);
    FN:=IncludeTrailingPathDelimiter(FCacheDir)+ADiscId;
    L.SaveToFile(FN);
  finally
    L.Free;
  end;
end;

Function TMainForm.LoadFromCache(ADiscID : Integer) : Boolean;

Var
  FN : String;

begin
  FN:=IncludeTrailingPathDelimiter(FCacheDir)+DiscIDtoStr(ADiscId);
  Result:=FileExists(FN);
  If Result then
    Result:=QueryFile(FN)>0;
end;


initialization
  {$I frmmain.lrs}

end.

